@@ -3,23 +3,34 @@ from cffi import FFI
 ffi = FFI()
 
 ffi.cdef("""
-// Type definitions
+// lber type definitions
+typedef ... BerElement;
+// Hack because cffi does not support #include in cdef.
+// These types are usually declared in lber_types.h
+typedef unsigned long ber_len_t;
+
+typedef struct berval {
+    ber_len_t bv_len;
+    char *bv_val;
+} BerValue, *BerVarray;
+
+
+// libldap type definitions
 typedef ... LDAP;
 typedef ... LDAPMessage;
 typedef ... LDAPControl;
-typedef ... BerElement;
-typedef ... berval;
 
 typedef struct ldapmod {
     int mod_op;
     char *mod_type;
     union {
         char **modv_strvals;
-    //    struct berval **modv_bvals;
+        struct berval **modv_bvals;
     } mod_vals;
 } LDAPMod;
 
 #define LDAP_VERSION3 ...
+
 #define LDAP_OPT_PROTOCOL_VERSION ...
 #define LDAP_OPT_X_TLS_REQUIRE_CERT ...
 #define LDAP_OPT_X_TLS_CACERTFILE ...
@@ -30,10 +41,16 @@ typedef struct ldapmod {
 #define LDAP_OPT_X_TLS_ALLOW ...
 #define LDAP_OPT_X_TLS_TRY ...
 #define LDAP_OPT_TIMELIMIT ...
+
+static char *const LDAP_ALL_USER_ATTRIBUTES;
+static char *const LDAP_ALL_OPERATIONAL_ATTRIBUTES;
+static char *const LDAP_NO_ATTRS;
+
 #define LDAP_NO_LIMIT ...
 #define LDAP_MOD_ADD ...
 #define LDAP_MOD_DELETE ...
 #define LDAP_MOD_REPLACE ...
+#define LDAP_MOD_BVALUES ...
 #define LDAP_SCOPE_BASE ...
 #define LDAP_SCOPE_ONELEVEL ...
 #define LDAP_SCOPE_SUBTREE ...
@@ -65,8 +82,8 @@ LDAPMessage *ldap_first_entry( LDAP *ld, LDAPMessage *result );
 LDAPMessage *ldap_next_entry( LDAP *ld, LDAPMessage *entry );
 
 // From ldap_get_values(3)
-char **ldap_get_values(LDAP *ld, LDAPMessage *entry, char *attr);
-int ldap_count_values(char **vals);
+struct berval **ldap_get_values_len(LDAP *ld, LDAPMessage *entry, char *attr);
+int ldap_count_values_len(struct berval **vals);
 
 // From ldap_get_dn(3)
 char *ldap_get_dn( LDAP *ld, LDAPMessage *entry );
@@ -110,6 +127,9 @@ char *ldap_err2string( int err );
 // From ldap_msgfree(3)
 int ldap_msgfree( LDAPMessage *msg );
 
+// From ldap_memfree(3)
+void ldap_memfree(void *p);
+
 int ldap_passwd_s(
         LDAP *ld,
         struct berval        *user,
@@ -121,6 +141,10 @@ int ldap_passwd_s(
 
 // From lber-types(3)
 struct berval *ber_bvstr(const char *str);
+void ber_bvfree(struct berval *bv);
+
+// From ber_free(3)
+void ber_free(BerElement *ber, int freebuf);
 
 """)
 
