@@ -5,6 +5,7 @@ from __future__ import unicode_literals
 from __future__ import print_function
 
 import copy
+import sys
 
 from ldapom.cdef import libldap, ffi
 from ldapom import attribute
@@ -88,7 +89,9 @@ class LDAPConnection(object):
 
     def __init__(self, uri, base, bind_dn, bind_password,
             cacertfile=None, require_cert=LDAP_OPT_X_TLS_NEVER,
-            timelimit=30, max_retry_reconnect=5):
+            timelimit=30, max_retry_reconnect=5,
+            schema_base="cn=subschema", enable_attribute_type_mapping=True,
+            retrieve_operational_attributes=False):
         """
         :param uri: URI of the server to connect to.
         :param base: Base DN for LDAP operations.
@@ -97,6 +100,10 @@ class LDAPConnection(object):
         :param cacertfile: If using SSL/TLS this is certificate of the server
         :param timelimit: Defines the time limit after which a search
             operation should be terminated by the server
+        :param schema_base: base DN for the schema description.
+        :param enable_attribute_type_mapping: Whether to enable the mapping of LDAP attribute types
+            to corresponding Python types. Requires the schema to be fetched when connecting. If
+            disabled, all attributes will be treated as a multi-value string attribute.
         """
         self._base = base
         self._uri = uri
@@ -106,6 +113,8 @@ class LDAPConnection(object):
         self._require_cert = require_cert
         self._max_retry_reconnect = max_retry_reconnect
         self._timelimit = timelimit
+        self._schema_base = schema_base
+        self._enable_attribute_type_mapping = enable_attribute_type_mapping
 
         self._connect()
 
@@ -142,16 +151,18 @@ class LDAPConnection(object):
                 compat._encode_utf8(self._bind_password))
         handle_ldap_error(err)
 
-        self._fetch_attribute_types()
+        if self._enable_attribute_type_mapping:
+            self._fetch_attribute_types()
 
     def _fetch_attribute_types(self):
+        attribute_type_definitions = attribute.DEFAULT_ATTRIBUTE_TYPES
         result = list(
-                self._raw_search(base="cn=subschema",
+                self._raw_search(base=self._schema_base,
                     scope=libldap.LDAP_SCOPE_BASE,
                     search_filter="(objectClass=*)",
                     retrieve_attributes=["attributeTypes"]))
         # Decode the type definitions returned to strings
-        attribute_type_definitions = map(compat._decode_utf8,
+        attribute_type_definitions += map(compat._decode_utf8,
                 result[0][1][compat._encode_utf8("attributeTypes")])
 
         self._attribute_types_by_name = attribute.build_attribute_types(
@@ -164,11 +175,21 @@ class LDAPConnection(object):
         :type name: str
         :rtype: a class object, a subclass of ``LDAPAttributeBase``.
         """
-        if name in self._attribute_types_by_name:
-            return self._attribute_types_by_name[name]
+        if self._enable_attribute_type_mapping:
+            if name in self._attribute_types_by_name:
+                return self._attribute_types_by_name[name]
+            else:
+                raise error.LDAPAttributeNameNotFoundError(
+                        'Attribute type "{}" not found.'.format(name))
         else:
-            raise error.LDAPAttributeNameNotFoundError(
-                    'Attribute type "{}" not found.'.format(name))
+            # Use a multi-value string attribute as the default
+            base_classes = [attribute.MultiValueAttributeMixin,
+                            attribute.BytesAttributeMixin,
+                            attribute.LDAPAttributeBase]
+            if sys.version_info[0] >= 3: # Python 3
+                return type("LDAPAttribute", tuple(base_classes), {})
+            else:
+                return type(bytes("LDAPAttribute"), tuple(base_classes), {})
 
     @_retry_reconnect
     def can_bind(self, bind_dn, bind_password):
@@ -188,7 +209,8 @@ class LDAPConnection(object):
         return True
 
     def _raw_search(self, search_filter=None, retrieve_attributes=None,
-            base=None, scope=libldap.LDAP_SCOPE_SUBTREE):
+            base=None, scope=libldap.LDAP_SCOPE_SUBTREE,
+            retrieve_operational_attributes=False):
         """
         Raw wrapper around OpenLDAP ldap_search_ext_s.
 
@@ -196,11 +218,13 @@ class LDAPConnection(object):
             if None is given.
         :type search_filter: List of str
         :param retrieve_attributes: List of attributes to retrieve. If None is
-            given, all are retrieved.
+            given, all user attributes are retrieved.
         :type retrieve_attributes: List of str
         :param base: Search base for the query.
         :type base: str
         :param scope: The search scope in the LDAP tree
+        :param retrieve_operational_attributes: Retrieve operational attributes of entries in
+            addition to user attributes if retrieve_attributes is not set.
         """
         search_result_p = ffi.new("LDAPMessage **")
 
@@ -208,15 +232,19 @@ class LDAPConnection(object):
         # still needed.
         prevent_garbage_collection = []
 
-        if retrieve_attributes is not None:
-            attrs_p = ffi.new("char*[{}]".format(len(retrieve_attributes) + 1))
-            for i, a in enumerate(retrieve_attributes):
-                attr_p = ffi.new("char[]", compat._encode_utf8(a))
-                prevent_garbage_collection.append(attr_p)
-                attrs_p[i] = attr_p
-            attrs_p[len(retrieve_attributes)] = ffi.NULL
-        else:
-            attrs_p = ffi.NULL
+        if retrieve_attributes is None:
+            retrieve_attributes = [
+                compat._decode_utf8(ffi.string(libldap.LDAP_ALL_USER_ATTRIBUTES))]
+            if retrieve_operational_attributes:
+                retrieve_attributes.append(
+                    compat._decode_utf8(ffi.string(libldap.LDAP_ALL_OPERATIONAL_ATTRIBUTES)))
+
+        attrs_p = ffi.new("char*[{}]".format(len(retrieve_attributes) + 1))
+        for i, a in enumerate(retrieve_attributes):
+            attr_p = ffi.new("char[]", compat._encode_utf8(a))
+            prevent_garbage_collection.append(attr_p)
+            attrs_p[i] = attr_p
+        attrs_p[len(retrieve_attributes)] = ffi.NULL
 
         err = libldap.ldap_search_ext_s(
                 self._ld,
@@ -245,33 +273,40 @@ class LDAPConnection(object):
                 current_attribute_str = ffi.string(current_attribute)
                 attribute_dict[current_attribute_str] = []
 
-                values_p = libldap.ldap_get_values(self._ld, current_entry,
+                values_p = libldap.ldap_get_values_len(self._ld, current_entry,
                         current_attribute)
-                for i in range(0, libldap.ldap_count_values(values_p)):
-                    attribute_dict[current_attribute_str].append(
-                            ffi.string(values_p[i]))
+                for i in range(0, libldap.ldap_count_values_len(values_p)):
+                    val = ffi.buffer(values_p[i].bv_val, values_p[i].bv_len)[:]
+                    attribute_dict[current_attribute_str].append(val)
 
+                libldap.ldap_memfree(current_attribute)
                 current_attribute = libldap.ldap_next_attribute(self._ld,
                         current_entry, ber_p[0])
-            # TODO: Call ber_free on ber_p[0]
+            libldap.ber_free(ber_p[0], 0)
 
             yield (dn, attribute_dict)
             current_entry = libldap.ldap_next_entry(self._ld, current_entry)
-            # TODO: Call ldap_msgfree on search_result
+
+        libldap.ldap_msgfree(search_result)
 
     @_retry_reconnect_generator
     def search(self, *args, **kwargs):
-        """Perform an LDAP search operation.
-
-        :rtype: Iterable of LDAPEntry.
-        """
+        """Perform an LDAP search operation."""
         return self._search(*args, **kwargs)
 
-    def _search(self, *args, **kwargs):
+    def _search(self, search_filter=None, retrieve_attributes=None,
+            base=None, scope=libldap.LDAP_SCOPE_SUBTREE,
+            retrieve_operational_attributes=False):
         """Search without retry_reconnect."""
         try:
-            for dn, attributes_dict in self._raw_search(*args, **kwargs):
-                entry = LDAPEntry(self, compat._decode_utf8(dn))
+            raw_search_result = self._raw_search(search_filter=search_filter,
+                                                 retrieve_attributes=retrieve_attributes,
+                                                 base=base,scope=scope,
+                                                 retrieve_operational_attributes=retrieve_operational_attributes)
+            for dn, attributes_dict in raw_search_result:
+                entry = LDAPEntry(self, compat._decode_utf8(dn),
+                                  retrieve_attributes=retrieve_attributes,
+                                  retrieve_operational_attributes=retrieve_operational_attributes)
                 entry._attributes = set()
                 for name, value in attributes_dict.items():
                     # TODO: Create the right type of LDAPAttribute here
@@ -280,6 +315,7 @@ class LDAPConnection(object):
                     attribute = attribute_type(compat._decode_utf8(name))
                     attribute._set_ldap_values(value)
                     entry._attributes.add(attribute)
+                entry._fetched_attributes = copy.deepcopy(entry._attributes)
                 yield entry
         except error.LDAPNoSuchObjectError:
             # If the search returned without results, "return" an empty generator.
@@ -379,21 +415,25 @@ class LDAPConnection(object):
             mod = ffi.new("LDAPMod *")
             prevent_garbage_collection.append(mod)
 
-            mod.mod_op = libldap.LDAP_MOD_REPLACE
+            mod.mod_op = libldap.LDAP_MOD_REPLACE | libldap.LDAP_MOD_BVALUES
 
             mod_type = ffi.new("char[]", compat._encode_utf8(attribute.name))
             prevent_garbage_collection.append(mod_type)
             mod.mod_type = mod_type
 
-            modv_strvals = ffi.new("char*[{}]".format(len(attribute._values) + 1))
-            prevent_garbage_collection.append(modv_strvals)
+            modv_bvals = ffi.new("BerValue*[{}]".format(len(attribute._values) + 1))
+            prevent_garbage_collection.append(modv_bvals)
             for j, value in enumerate(attribute._get_ldap_values()):
-                strval = ffi.new("char[]", value)
-                prevent_garbage_collection.append(strval)
-                modv_strvals[j] = strval
-            modv_strvals[len(attribute._values)] = ffi.NULL
-            mod.mod_vals = {"modv_strvals": modv_strvals}
-
+                modv_berval = ffi.new("BerValue *")
+                prevent_garbage_collection.append(modv_berval)
+                modv_berval.bv_len = len(value)
+                bval = ffi.new("char[]", len(value))
+                prevent_garbage_collection.append(bval)
+                ffi.buffer(bval)[:] = value
+                modv_berval.bv_val = bval
+                modv_bvals[j] = modv_berval
+            modv_bvals[len(attribute._values)] = ffi.NULL
+            mod.mod_vals = {"modv_bvals": modv_bvals}
             mods[i] = mod
         mods[len(changed_attributes)] = ffi.NULL
 
@@ -412,15 +452,15 @@ class LDAPConnection(object):
         entry._fetched_attributes = copy.deepcopy(entry._attributes)
 
     @_retry_reconnect
-    def fetch(self, entry):
+    def fetch(self, entry, *args, **kwargs):
         """Fetch an entry's attributes from the LDAP server.
 
         :param entry: The entry to fetch.
         :type entry: ldapom.LDAPEntry
         """
         try:
-            fetched_entry = next(self._search(base=entry.dn,
-                scope=libldap.LDAP_SCOPE_BASE))
+            fetched_entry = next(self._search(*args, base=entry.dn,
+                scope=libldap.LDAP_SCOPE_BASE, **kwargs))
             entry._attributes = fetched_entry._attributes
             entry._fetched_attributes = copy.deepcopy(entry._attributes)
         except StopIteration:
